﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace moxa
{
    public class mxdevice
    {
        #region Win32 API Constant
        public readonly static IntPtr INVALID_HANDLE_VALUE = new IntPtr(-1);
        public const UInt32 OPEN_EXISTING = 3;
        public const UInt32 GENERIC_READ = 0x80000000;
        public const UInt32 GENERIC_WRITE = 0x40000000;
        public const UInt32 FILE_SHARE_READ = 0x00000001;
        public const UInt32 FILE_SHARE_WRITE = 0x00000002;

        public const UInt32 GPD_TYPE = 40000;
        #endregion

        #region Device IO Control Code
        public const UInt32 FILE_DEVICE_HAL = 0x00000101;
        public const UInt32 FILE_DEVICE_CONSOLE = 0x00000102;
        public const UInt32 FILE_DEVICE_PSL = 0x00000103;
        public const UInt32 FILE_DEVICE_SERVICE = 0x00000104;
        public const UInt32 FILE_DEVICE_NLED = 0x00000105;
        public const UInt32 FILE_DEVICE_NOTIFY = 0x00000106;

        public const UInt32 METHOD_BUFFERED = 0;
        public const UInt32 METHOD_IN_DIRECT = 1;
        public const UInt32 METHOD_OUT_DIRECT = 2;
        public const UInt32 METHOD_NEITHER = 3;

        public const UInt32 FILE_ANY_ACCESS = 0;
        public const UInt32 FILE_READ_ACCESS = (0x0001);    // file & pipe
        public const UInt32 FILE_WRITE_ACCESS = (0x0002);    // file & pipe

        public static uint IOCTL_SET_DOUT = CTL_CODE(GPD_TYPE, 0x900, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        public static uint IOCTL_GET_DOUT = CTL_CODE(GPD_TYPE, 0x901, METHOD_BUFFERED, FILE_READ_ACCESS);
        public static uint IOCTL_GET_DIN = CTL_CODE(GPD_TYPE, 0x902, METHOD_BUFFERED, FILE_READ_ACCESS);
        public static uint IOCTL_GET_DIN_ALL = CTL_CODE(GPD_TYPE, 0x903, METHOD_BUFFERED, FILE_READ_ACCESS);
        public static uint IOCTL_GET_DOUT_ALL = CTL_CODE(GPD_TYPE, 0x904, METHOD_BUFFERED, FILE_READ_ACCESS);

        public const uint MOXA_SET_OP_MODE = (0x400 + 66);
        public const uint MOXA_GET_OP_MODE = (0x400 + 67);
        public const uint RS232_MODE = 0;
        public const uint RS485_2WIRE_MODE = 1;
        public const uint RS422_MODE = 2;
        public const uint RS485_4WIRE_MODE = 3;

        #endregion

        #region Win32 API imports
        [DllImport("kernel32.dll", EntryPoint = "Sleep", SetLastError = true)]
        private static extern void Sleep(UInt32 dwMilliseconds);

        [DllImport("kernel32.dll", EntryPoint = "CreateFile", SetLastError = true)]
        private static extern IntPtr CreateFile(String lpFileName, UInt32 dwDesiredAccess, UInt32 dwShareMode, IntPtr lpSecurityAttributes, UInt32 dwCreationDisposition, UInt32 dwFlagsAndAttributes, IntPtr hTemplateFile);

        [DllImport("kernel32.dll", EntryPoint = "CloseHandle", SetLastError = true)]
        private static extern int CloseHandle(IntPtr hObject);

        [DllImport("kernel32.dll", EntryPoint = "WriteFile", SetLastError = true)]
        private static extern int WriteFile(IntPtr hFile, byte[] lpBuffer, Int32 nNumberOfBytesToRead, out Int32 lpNumberOfBytesRead, IntPtr lpOverlapped);

        [DllImport("kernel32.dll", EntryPoint = "ReadFile", SetLastError = true)]
        private static extern int ReadFile(IntPtr hFile, byte[] lpBuffer, Int32 nNumberOfBytesToRead, ref Int32 lpNumberOfBytesRead, IntPtr lpOverlapped);

        [DllImport("kernel32.dll", EntryPoint = "DeviceIoControl", SetLastError = true)]
        public static extern bool DeviceIoControl_Int(
            IntPtr hDevice,
            uint dwIoControlCode,
            IntPtr lpInBuffer,
            int nInBufferSize,
            IntPtr lpOutBuffer,
            int nOutBufferSize,
            ref int lpBytesReturned,
            IntPtr lpOverlapped
            );

        [DllImport("kernel32.dll", EntryPoint = "DeviceIoControl", SetLastError = true)]
        public static extern bool DeviceIoControl_Int2(
            IntPtr hDevice,
            uint dwIoControlCode,
            IntPtr lpInBuffer,
            int nInBufferSize,
            ref uint lpOutBuffer,
            int nOutBufferSize,
            ref int lpBytesReturned,
            IntPtr lpOverlapped
            );

        [DllImport("kernel32.dll", EntryPoint = "DeviceIoControl", SetLastError = true)]
        public static extern bool DeviceIoControl_Byte(
            IntPtr hDevice,
            uint dwIoControlCode,
            ref byte lpInBuffer,
            int nInBufferSize,
            ref byte lpOutBuffer,
            int nOutBufferSize,
            ref int lpBytesReturned,
            IntPtr lpOverlapped
            );
        [DllImport("kernel32.dll", EntryPoint = "DeviceIoControl", SetLastError = true)]
        public static extern bool DeviceIoControl_DWORD(
            IntPtr hDevice,
            uint dwIoControlCode,
            ref uint lpInBuffer,
            int nInBufferSize,
            ref uint lpOutBuffer,
            int nOutBufferSize,
            ref int lpBytesReturned,
            IntPtr lpOverlapped
            );
        #endregion

        #region Structures Definition
        [StructLayout(LayoutKind.Sequential)]
        public struct _DIO
        {
            [MarshalAs(UnmanagedType.I4)]
            public int port;
            [MarshalAs(UnmanagedType.I4)]
            public int data;
        };

        [StructLayout(LayoutKind.Sequential)]
        public struct SYSTEMTIME
        {
            [MarshalAs(UnmanagedType.U2)]
            public short Year;
            [MarshalAs(UnmanagedType.U2)]
            public short Month;
            [MarshalAs(UnmanagedType.U2)]
            public short DayOfWeek;
            [MarshalAs(UnmanagedType.U2)]
            public short Day;
            [MarshalAs(UnmanagedType.U2)]
            public short Hour;
            [MarshalAs(UnmanagedType.U2)]
            public short Minute;
            [MarshalAs(UnmanagedType.U2)]
            public short Second;
            [MarshalAs(UnmanagedType.U2)]
            public short Milliseconds;
        }
        #endregion

        private String devFile;
        private IntPtr devHndl = INVALID_HANDLE_VALUE;

        public mxdevice(String devFile)
        {
            this.devFile = devFile;
        }

        /* open for read only */
        public bool OpenReadOnly()
        {
            return OpenX(GENERIC_READ);
        }

        /* open for write only */
        public bool OpenWriteOnly()
        {
            return OpenX(GENERIC_WRITE);
        }

        /* open for read/write */
        public bool Open()
        {
            return OpenX(GENERIC_WRITE | GENERIC_READ);
        }

        public bool isOpen()
        {
            return (devHndl != INVALID_HANDLE_VALUE);
        }

        public void Close()
        {
            if (devHndl != INVALID_HANDLE_VALUE)
            {
                CloseHandle(devHndl);
                devHndl = INVALID_HANDLE_VALUE;
            }
        }

        public int ReadDev(byte[] inbytes, Int32 nReadBufLen, ref int nRetRead)
        {
            return ReadFile(devHndl, inbytes, nReadBufLen, ref nRetRead, IntPtr.Zero);
        }

        public int WriteDev(byte[] inbytes, Int32 nWriteLen, out Int32 nRetWrite)
        {
            return WriteFile(devHndl, inbytes, nWriteLen, out nRetWrite, IntPtr.Zero);
        }

        public bool IoControl(uint code)
        {
            int nBytesReturned = 0;
            byte bIn = 0, bOut = 0;
            return DeviceIoControl_Byte(devHndl, code, ref bIn, 0, ref bOut, 0, ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint code, ref byte input)
        {
            int nBytesReturned = 0;
            byte bOut = 0;
            return DeviceIoControl_Byte(devHndl, code, ref input, sizeof(byte), ref bOut, 0, ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint code, IntPtr pInput)
        {
            int nBytesReturned = 0;
            IntPtr pOutput = new IntPtr(0);
            return DeviceIoControl_Int(devHndl, code, pInput, sizeof(uint), pOutput, 0, ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint code, ref byte input, ref byte output)
        {
            int nBytesReturned = 0;
            return DeviceIoControl_Byte(devHndl, code, ref input, sizeof(byte), ref output, sizeof(byte), ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint code, IntPtr pInput, IntPtr pOutput)
        {
            int nBytesReturned = 0;
            return DeviceIoControl_Int(devHndl, code, pInput, sizeof(uint), pOutput, sizeof(uint), ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint code, IntPtr pInput, int nInSize, IntPtr pOutput, int nOutSize)
        {
            int nBytesReturned = 0;
            return DeviceIoControl_Int(devHndl, code, pInput, nInSize, pOutput, nOutSize, ref nBytesReturned, IntPtr.Zero);
        }

        public bool IoControl(uint dwIoControlCode, ref uint lpInBuffer, int nInBufferSize, ref uint lpOutBuffer, int nOutBufferSize)
        {
            int nBytesReturned = 0;
            return DeviceIoControl_DWORD(devHndl, dwIoControlCode, ref lpInBuffer, nInBufferSize, ref lpOutBuffer, nOutBufferSize, ref nBytesReturned, IntPtr.Zero);
        }

        public bool DeviceIoControl(uint dwIoControlCode, IntPtr lpInBuffer, int nInBufferSize, IntPtr lpOutBuffer, int nOutBufferSize, ref int lpBytesReturned, IntPtr lpOverlapped)
        {
            return DeviceIoControl_Int(devHndl, dwIoControlCode, lpInBuffer, nInBufferSize, lpOutBuffer, nOutBufferSize, ref lpBytesReturned, lpOutBuffer);
        }

        public bool DeviceIoControl(uint dwIoControlCode, IntPtr lpInBuffer, int nInBufferSize, ref uint lpOutBuffer, int nOutBufferSize, ref int lpBytesReturned, IntPtr lpOverlapped)
        {
            return DeviceIoControl_Int2(devHndl, dwIoControlCode, lpInBuffer, nInBufferSize, ref lpOutBuffer, nOutBufferSize, ref lpBytesReturned, lpOverlapped);
        }

        /* generic device file open */
        private bool OpenX(UInt32 flag)
        {
            if (devHndl != INVALID_HANDLE_VALUE)
                return false;
            devHndl = CreateFile(devFile, flag, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
            if (devHndl == INVALID_HANDLE_VALUE)
                return false;
            return true;
        }

        private static uint CTL_CODE(uint DeviceType, uint Function, uint Method, uint Access)
        {
            return ((DeviceType << 16) | (Access << 14) | (Function << 2) | Method);
        }
    }
}
